"""Token"""
import json
from typing import Any

from loguru import logger

from . import cipher, utils


def generate_token(key: str = 'Fc0zXCmGKd7tPu6W', timeout: int = 3600, data: Any = None) -> (str | None):
    try:
        now = utils.datetime_now()

        if now is None:
            return None

        offset = utils.datetime_offset(now, seconds=+timeout)

        if offset is None:
            return None

        source = json.dumps(
            obj={
                "datetime": utils.datetime_to_string(offset),
                "data": data
            },
            default=str
        )

        aes_cipher = cipher.AESCipher(key=key, algorithm='sha256')

        return aes_cipher.encrypt(source)

    except Exception as e:
        logger.exception(e)
        return None


def parsing_token(token_string: str, key: str = 'Fc0zXCmGKd7tPu6W') -> (dict | None):
    try:
        if not utils.isTrue(token_string, str):
            return None

        aes_cipher = cipher.AESCipher(key=key, algorithm='sha256')

        target = aes_cipher.decrypt(token_string)

        if target is None:
            return None

        source: dict = json.loads(target)

        source['datetime'] = utils.datetime_string_to_datetime(source['datetime'])

        return source

    except Exception as e:
        logger.exception(e)
        return None


def certify_token(token_string: str, key: str = 'Fc0zXCmGKd7tPu6W') -> bool:
    try:

        result = parsing_token(token_string, key)

        if result is None:
            return False

        if result.get('datetime') < utils.datetime_now():  # type: ignore
            return False

        return True

    except Exception as e:
        logger.exception(e)
        return False



"""Cipher"""
# https://docs.python.org/3.10/library/hashlib.html
# https://www.pycrypto.org/
# https://stackoverflow.com/a/21928790
# pip install pycryptodome
import base64
import hashlib

from Crypto import Random
from Crypto.Cipher import AES
from loguru import logger


class AESCipher:
    """AESCipher"""

    def __init__(self, key: str = 'vB7DoRm9C2Kd', algorithm: str = 'sha256'):

        self.bs = AES.block_size

        # dir(hashlib)
        match True:
            case True if algorithm == 'md5':
                self.key = hashlib.md5(key.encode()).digest()
            case True if algorithm == 'sha1':
                self.key = hashlib.sha1(key.encode()).digest()
            case True if algorithm == 'sha224':
                self.key = hashlib.sha224(key.encode()).digest()
            case True if algorithm == 'sha256':
                self.key = hashlib.sha256(key.encode()).digest()
            case True if algorithm == 'sha384':
                self.key = hashlib.sha384(key.encode()).digest()
            case True if algorithm == 'sha512':
                self.key = hashlib.sha512(key.encode()).digest()
            case True if algorithm == 'sha3_224':
                self.key = hashlib.sha3_224(key.encode()).digest()
            case True if algorithm == 'sha3_256':
                self.key = hashlib.sha3_256(key.encode()).digest()
            case True if algorithm == 'sha3_384':
                self.key = hashlib.sha3_384(key.encode()).digest()
            case True if algorithm == 'sha3_512':
                self.key = hashlib.sha3_512(key.encode()).digest()
            # case True if algorithm == 'shake_128':
            #     self.key = hashlib.shake_128(key.encode()).digest()
            # case True if algorithm == 'shake_256':
            #     self.key = hashlib.shake_256(key.encode()).digest()
            case _:
                self.key = hashlib.sha256(key.encode()).digest()

    def encrypt(self, raw: str) -> str | None:
        try:
            raw = self._pad(raw)
            iv = Random.new().read(AES.block_size)
            cipher = AES.new(self.key, AES.MODE_CBC, iv)
            return base64.b64encode(iv + cipher.encrypt(raw.encode())).decode('utf-8')
        except Exception as e:
            logger.exception(e)
            return None

    def decrypt(self, enc: str) -> str | None:
        try:
            enc_bytes = base64.b64decode(enc)
            iv = enc_bytes[:AES.block_size]
            cipher = AES.new(self.key, AES.MODE_CBC, iv)
            return self._unpad(cipher.decrypt(enc_bytes[AES.block_size:])).decode('utf-8')
        except Exception as e:
            logger.exception(e)
            return None

    def _pad(self, s):
        return s + (self.bs - len(s) % self.bs) * chr(self.bs - len(s) % self.bs)

    @staticmethod
    def _unpad(s: bytes) -> bytes:
        return s[:-ord(s[len(s) - 1:])]



"""发送邮件"""
# https://stackoverflow.com/questions/882712/sending-html-email-using-python
import smtplib
from email.header import Header
from email.mime.image import MIMEImage
from email.mime.multipart import MIMEMultipart
from email.mime.text import MIMEText
from email.utils import formataddr, parseaddr
from typing import Optional, TypedDict, cast

from loguru import logger

from . import utils


class TypedSMTP(TypedDict):
    """smtp type"""
    server: str
    port: int
    tls: bool

class TypedSender(TypedDict):
    """sender type"""
    name: str
    address: str
    password: str

class TypedBody(TypedDict, total=False):
    """body type"""
    content: str
    type: Optional[str]

def format_parse(s):
    _name, _addr = parseaddr(s)
    return formataddr((Header(_name, "utf-8").encode(), _addr))

def sendemail(
    smtp: TypedSMTP,
    sender: TypedSender,
    recipients: str | list,
    subject: str,
    body: TypedBody,
    images: None | list = None
) -> bool:
    """
    smtp SMTP信息

        server  SMTP地址
        port    SMTP端口
        tls     是否使用TLS

    sender 发件人信息

        name     发件人名称
        address  发件人邮箱地址
        password 发件人邮箱密码(SMTP)

    recipients  收件人(或列表)

    subject     邮件主题

    body        邮件主体

        content 内容
        type    类型 (默认 plain, 或者 file, 或者 html)

    images 图片列表(可选)

        cid  图片CID
        path 图片路径
    """

    logger.success("sendemail start")

    try:

        # 邮件主体

        if not utils.isTrue(body, dict):
            logger.error("body error")
            return False

        message: MIMEMultipart = MIMEMultipart()

        body_content = cast(str, body.get("content"))

        if not utils.isTrue(body_content, str):
            logger.error(f"body content error: {body_content}")
            return False

        body_type = cast(str, body.get("type"))

        # 从文本文件读取内容
        if body_type == "file":
            with open(body_content, "r", encoding="utf-8") as file:
                message.attach(MIMEText(file.read(), "plain", "utf-8"))

        # 从HTML文件读取内容
        elif body_type == "html":
            message = MIMEMultipart("related")
            with open(body_content, "r", encoding="utf-8") as file:
                message.attach(MIMEText(file.read(), "html", "utf-8"))

        # 纯文本内容
        else:
            message.attach(MIMEText(body_content, "plain", "utf-8"))

        # ------------------------------------------------------------------------------------------

        # SMTP

        if not utils.isTrue(smtp, dict):
            logger.error("smtp error")
            return False

        smtp_host = cast(str, smtp.get("server"))
        smtp_port = cast(int, smtp.get("port"))
        smtp_tls = cast(bool, smtp.get("tls"))

        if not utils.isTrue(smtp_host, str):
            logger.error(f"smtp host error: {smtp_host}")
            return False

        if not utils.isTrue(smtp_port, int):
            logger.error(f"smtp port error: {smtp_port}")
            return False

        smtp_tls = utils.isTrue(smtp_tls, bool)

        # ------------------------------------------------------------------------------------------

        # 发件人信息

        if not utils.isTrue(sender, dict):
            logger.error("sender error")
            return False

        sender_name = cast(str, sender.get("name"))
        sender_address = cast(str, sender.get("address"))
        sender_password = cast(str, sender.get("password"))

        if not utils.isTrue(sender_name, str):
            logger.error(f"sender name error: {sender_name}")
            return False

        if not utils.isTrue(sender_address, str):
            logger.error(f"sender address error: {sender_address}")
            return False

        if not utils.isTrue(sender_password, str):
            logger.error(f"sender password error: {sender_password}")
            return False

        message["From"] = formataddr((sender_name, sender_address))

        # ------------------------------------------------------------------------------------------

        # 收件人(或列表)

        if utils.isTrue(recipients, str):
            message["To"] = format_parse(recipients)
        elif utils.isTrue(recipients, list):
            message["To"] = ", ".join(list(map(format_parse, recipients)))
        else:
            logger.error("recipients error")
            return False

        # ------------------------------------------------------------------------------------------

        # 邮件主题

        if not utils.isTrue(subject, str):
            logger.error("subject error")
            return False

        message["Subject"] = subject

        # ------------------------------------------------------------------------------------------

        if images is not None and utils.isTrue(images, list):

            for image in images:

                try:

                    if utils.check_file_type(image.get("path", ""), "file"):

                        # 添加图片
                        # with open(image_path, "rb") as image_file:
                        #     mime_image = MIMEImage(image_file.read())
                        #     # Define the image's ID as referenced above
                        #     mime_image.add_header("Content-ID", "<CID>")
                        #     message.attach(mime_image)

                        with open(image["path"], "rb") as _image_file:
                            mime_image = MIMEImage(_image_file.read())
                            mime_image.add_header("Content-ID", f"<{image['cid']}>")
                            message.attach(mime_image)

                except Exception as e:
                    logger.exception(e)
                    return False

        # ------------------------------------------------------------------------------------------

        # 发送邮件

        # SMTP.sendmail(from_addr, to_addrs, msg, mail_options=(), rcpt_options=())
        #
        #     to_addrs = sender_to + sender_cc
        #     https://docs.python.org/3/library/smtplib.html#smtplib.SMTP.sendmail
        #     https://gist.github.com/AO8/c5a6f747eeeca02351152ae8dc79b537

        # if smtp.get('ssl', False) is True:

        #     with smtplib.SMTP_SSL(smtp_host, smtp_port) as _smtp:
        #         _smtp.login(sender_address, sender_password)
        #         _smtp.sendmail(sender_address, recipients, _message.as_string())

        # else:

        with smtplib.SMTP(smtp_host, smtp_port) as smtp_server:
            if smtp_tls is True:
                smtp_server.starttls()
            smtp_server.login(sender_address, sender_password)
            smtp_server.sendmail(sender_address, recipients, message.as_string())

            logger.success("sendemail success")

            return True

    except Exception as e:

        # 忽略腾讯邮箱返回的异常
        if e.args == (-1, b'\x00\x00\x00'):
            # pass
            return True

        logger.error('sendemail error')
        logger.exception(e)
        return False



"""Bottle Extensions"""
from typing import Callable

from . import bottle, utils


def enable_cors(fn: Callable) -> Callable:
    """Bottle CORS"""
    # 参考文档:
    # - https://stackoverflow.com/a/17262900
    # - https://developer.mozilla.org/en-US/docs/Web/HTTP/Headers/Access-Control-Allow-Headers
    # - https://developer.mozilla.org/en-US/docs/Web/HTTP/Headers/Access-Control-Allow-Methods
    # - https://developer.mozilla.org/en-US/docs/Web/HTTP/Headers/Access-Control-Allow-Origin
    def cors(*args, **kwargs):
        bottle.response.headers['Access-Control-Allow-Headers'] = '*'
        bottle.response.headers['Access-Control-Allow-Methods'] = '*'
        bottle.response.headers['Access-Control-Allow-Origin'] = '*'
        if bottle.request.method != 'OPTIONS':
            return fn(*args, **kwargs)
        return None
    return cors

def request_json() -> bottle.DictProperty | None:
    """Bottle Request JSON"""
    try:
        data = bottle.request.json
        if utils.isTrue(data, dict):
            return data
        return None
    except Exception as e:
        print(e)
        return None



"""企业微信"""
# 企业微信开发者中心
#     https://developer.work.weixin.qq.com/
#     https://developer.work.weixin.qq.com/document/path/90313 (全局错误码)
# 参考文档:
#     https://www.gaoyuanqi.cn/python-yingyong-qiyewx/
#     https://www.jianshu.com/p/020709b130d3
import json
import time

import requests
from loguru import logger

from . import utils


class QYWX:
    """企业微信"""

    url_prefix = "https://qyapi.weixin.qq.com"
    work_id: str | None = None
    agent_id: str | None = None
    agent_secret: str | None = None
    access_token: str | None = None

    def __init__(self, work_id: str | None, agent_id: str | None, agent_secret: str | None):
        """Initiation"""
        self.work_id = work_id
        self.agent_id = agent_id
        self.agent_secret = agent_secret

        """获取 Token"""
        self.getaccess_token()

    def getaccess_token(self) -> str | None:
        try:
            response = requests.get(f"{self.url_prefix}/cgi-bin/gettoken?corpid={self.work_id}&corpsecret={self.agent_secret}", timeout=10)

            if response.status_code != 200:
                self.access_token = None
                return None

            result: dict = response.json()
            self.access_token = result.get('access_token')
            return result.get('access_token')

        except Exception as e:
            logger.exception(e)
            return None

    def get_agent_list(self) -> dict | str | None:
        try:
            if self.access_token is None:
                self.getaccess_token()
            response = requests.get(f"{self.url_prefix}/cgi-bin/agent/list?access_token={self.access_token}", timeout=10)
            if response.status_code == 200:
                response_data: dict = response.json()
                if response_data.get('errcode') == 42001:
                    self.getaccess_token()
                    time.sleep(1)
                    self.get_agent_list()
                return response_data
            return response.text
        except Exception as e:
            logger.exception(e)
            return None

    def get_department_list(self, eid: str | None = None) -> dict | str | None:
        """eid: Enterprise ID"""
        try:
            if self.access_token is None:
                self.getaccess_token()
            response = requests.get(f"{self.url_prefix}/cgi-bin/department/list?access_token={self.access_token}&id={eid}", timeout=10)
            if response.status_code == 200:
                response_data: dict = response.json()
                if response_data.get('errcode') == 42001:
                    self.getaccess_token()
                    time.sleep(1)
                    self.get_department_list(eid)
                return response_data
            return response.text
        except Exception as e:
            logger.exception(e)
            return None

    def get_user_list(self, eid: str | None = None) -> dict | str | None:
        """eid: Enterprise ID"""
        try:
            if self.access_token is None:
                self.getaccess_token()
            response = requests.get(f"{self.url_prefix}/cgi-bin/user/list?access_token={self.access_token}&department_id={eid}", timeout=10)
            if response.status_code == 200:
                response_data: dict = response.json()
                if response_data.get('errcode') == 42001:
                    self.getaccess_token()
                    time.sleep(1)
                    self.get_user_list(eid)
                return response_data
            return response.text
        except Exception as e:
            logger.exception(e)
            return None

    def get_user_id_by_mobile(self, mobile: str | None = None) -> dict | str | None:
        try:
            if self.access_token is None:
                self.getaccess_token()
            json_string = json.dumps({'mobile': mobile})
            response = requests.post(f"{self.url_prefix}/cgi-bin/user/getuserid?access_token={self.access_token}", data=json_string, timeout=10)
            if response.status_code == 200:
                response_data: dict = response.json()
                if response_data.get('errcode') == 42001:
                    self.getaccess_token()
                    time.sleep(1)
                    self.get_user_id_by_mobile(mobile)
                return response_data
            return response.text
        except Exception as e:
            logger.exception(e)
            return None

    def get_user_info(self, eid: str | None = None) -> dict | str | None:
        """eid: Enterprise ID"""
        try:
            if self.access_token is None:
                self.getaccess_token()
            response = requests.get(f"{self.url_prefix}/cgi-bin/user/get?access_token={self.access_token}&userid={eid}", timeout=10)
            if response.status_code == 200:
                response_data: dict = response.json()
                if response_data.get('errcode') == 42001:
                    self.getaccess_token()
                    time.sleep(1)
                    self.get_user_info(eid)
                return response_data
            return response.text
        except Exception as e:
            logger.exception(e)
            return None

    def send_message_by_mobile(self, mobile: str | list, message: str) -> bool:
        """发送消息"""

        # 参考文档:
        # https://developer.work.weixin.qq.com/document/path/90235

        try:
            if self.access_token is None:
                self.getaccess_token()

            users: list = []

            match True:
                case True if isinstance(mobile, list) and utils.isTrue(mobile, list):
                    users = mobile
                case True if isinstance(mobile, str) and utils.isTrue(mobile, str):
                    users.append(mobile)
                case _:
                    return False

            for user in users:
                user_object = self.get_user_id_by_mobile(user)

                if not isinstance(user_object, dict):
                    continue

                json_dict = {
                    'touser': user_object.get('userid'),
                    'msgtype': 'text',
                    'agentid': self.agent_id,
                    'text': {'content': message},
                    'safe': 0,
                    'enable_id_trans': 0,
                    'enable_duplicate_check': 0,
                    'duplicate_check_interval': 1800
                }
                json_string = json.dumps(json_dict)
                response = requests.post(f"{self.url_prefix}/cgi-bin/message/send?access_token={self.access_token}", data=json_string, timeout=10)
                if response.status_code == 200:
                    response_data: dict = response.json()
                    if response_data.get('errcode') == 42001:
                        self.getaccess_token()
                        time.sleep(1)
                        self.send_message_by_mobile(mobile, message)

            return True

        except Exception as e:
            logger.exception(e)
            return False



"""FTP"""
# ftplib: https://docs.python.org/3.11/library/ftplib.html
import os
from ftplib import FTP
from pathlib import Path

from loguru import logger


class XFTP:
    """XFTP"""

    def __init__(self, host="127.0.0.1", port=21, username="anonymous", password="", encoding="UTF-8", debuglevel=0):
        """Initiation"""
        self.ftp = FTP()
        self.ftp.set_debuglevel(debuglevel)
        self.host = host
        self.port = port
        self.username = username
        self.password = password
        self.encoding = encoding
        self.retry = 1

    def connect(self) -> bool:
        """FTP connect"""
        try:
            self.ftp.connect(host=self.host, port=self.port, timeout=10)
            self.ftp.encoding = self.encoding
            self.ftp.login(user=self.username, passwd=self.password)
            logger.success("FTP connect success")
            logger.info("-" * 80)
            return True
        except Exception as e:
            # print(f"FTP connect error: {e}, retry...")
            # if self.retry >= 3:
            #     print("FTP connect faild")
            #     return False
            # self.retry += 1
            # self.connect()
            logger.exception(e)
            return False

    def close(self, info=None) -> bool:
        """FTP close"""
        if info is not None:
            logger.info(info)
        try:
            self.ftp.quit()
        except Exception as e:
            logger.exception(e)
            self.ftp.close()
        logger.info("-" * 80)
        logger.success("FTP connect closed")
        return True

    def get_file_list(self, target='/') -> list[str] | None:
        """Get file list"""
        try:
            self.chdir_to_remote(target)
            return self.ftp.nlst()
        except Exception as e:
            logger.exception(e)
            return None

    def get_file_size(self, file, target="/") -> int | None:
        """Get file size"""
        try:
            self.chdir_to_remote(target)
            return self.ftp.size(file)
        except Exception as e:
            logger.exception(e)
            return None

    def mkdir(self, target="/") -> bool:
        """创建目录 (从 / 目录依次递增创建子目录. 如果目录存在, 创建目录时会报错, 所以这里忽略所有错误.)"""
        try:
            dir_list = target.split("/")
            for i, _ in enumerate(dir_list):
                dir_path = "/".join(dir_list[:i + 1])
                try:
                    self.ftp.mkd(dir_path)
                except Exception as e:
                    logger.exception(e)
            return True
        except Exception as e:
            logger.exception(e)
            return False

    def chdir_to_remote(self, target="/") -> bool:
        """change to remote directory"""
        try:
            self.ftp.cwd(target)
            return True
        except Exception as e:
            self.close(f"remote directory error: {target}")
            logger.exception(e)
            return False

    def x_exit(self, info=None):
        """Exit"""
        if info is not None:
            logger.info(info)
        # 注意: exit() 并不会退出脚本, 配合 try 使用
        exit()

    def x_exec(self, local_dir=".", local_file="", remote_dir="/", remote_file="", upload=False):
        """Download or Upload"""

        bufsize = 1024
        local_path = f"{local_dir}/{local_file}"
        remote_path = f"{remote_dir}/{remote_file}"

        info = 'Download'
        if upload is True:
            info = 'Upload'

        # 检查参数
        if upload is True:
            if local_file == "":
                self.close('Argument Miss: local file')
            # 如果没有设置 远程文件 名称, 则使用 本地文件 名称
            if remote_file == "":
                remote_file = local_file
                remote_path = f"{remote_dir}/{remote_file}"
        else:
            if remote_file == "":
                self.close("Argument Miss: remote file")
            # 如果没有设置 本地文件 名称, 则使用 远程文件 名称
            if local_file == "":
                local_file = remote_file
                local_path = f"{local_dir}/{local_file}"

        # 进入本地目录
        try:
            if upload is True:
                # 检查本地目录
                stat = Path(local_dir)
                if stat.exists() is False:
                    self.close(f"Local directory error: {local_dir}")
            else:
                # 创建本地目录
                Path(local_dir).mkdir(parents=True, exist_ok=True)
            # 进入本地目录
            os.chdir(local_dir)
        except Exception as e:
            logger.exception(e)
            # 第一层 try 使用 self.x_exit() 无效, 直接使用 self.close()
            self.close(f"Local directory error: {local_dir}")

        # 上传或下载
        try:

            if upload is True:

                # 上传

                # 创建远程目录
                if remote_dir != "/":
                    self.mkdir(remote_dir)

                # 进入远程目录
                self.chdir_to_remote(remote_dir)

                # 上传文件
                stat = Path(local_file)
                if stat.exists() and stat.is_file():
                    with open(local_file, "rb") as fid:
                        self.ftp.storbinary(f'STOR {remote_file}', fid, bufsize)
                    logger.success(f"{info} success: {local_path.replace('//', '/')} -> {remote_path.replace('//', '/')}")
                    return True

                self.x_exit(f"{info} error: {local_path.replace('//', '/')} is not exist")

            else:

                # 下载

                # 进入远程目录
                self.chdir_to_remote(remote_dir)

                # 下载文件
                if remote_file in self.ftp.nlst():
                    with open(local_file, "wb") as fid:
                        self.ftp.retrbinary(f'RETR {remote_file}', fid.write, bufsize)
                    logger.success(f"{info} success: {remote_path.replace('//', '/')} -> {local_path.replace('//', '/')}")
                    return True

                self.x_exit(f"{info} error: {remote_path.replace('//', '/')} is not exist")

        except Exception as e:
            # 第一层 try 使用 self.x_exit() 无效, 直接使用 self.close()
            # self.close('{} faild! Please check {} or {}'.format(info, local_path, remote_path))
            self.close(f"{info} error: {e}")
            return False

    def handle_all(self, local_dir=".", remote_dir="/", upload=False):
        """Handle All"""
        if upload is True:
            # 检查本地目录
            stat = Path(local_dir)
            if stat.exists() is False:
                self.close(f"Local directory error: {local_dir}")
            # 获取文件列表
            local_files = [f for f in os.listdir(local_dir) if os.path.isfile(os.path.join(local_dir, f))]
            for i in local_files:
                self.x_exec(local_dir=local_dir, remote_dir=remote_dir, local_file=i, upload=True)
        else:
            remote_files = self.get_file_list(remote_dir)
            if remote_files is not None:
                for i in remote_files:
                    self.x_exec(local_dir=local_dir, remote_dir=remote_dir, remote_file=i)

    def retrlines(self, remote_dir="/", cmd="LIST"):
        """Retrlines"""
        try:
            self.chdir_to_remote(remote_dir)
            print(self.ftp.retrlines(cmd))
            self.close()
        except Exception as e:
            # 第一层 try 使用 self.x_exit() 无效, 直接使用 self.close()
            self.close(e)



"""Utils"""
import csv
import datetime
import hashlib
import json
import os
import subprocess
import time
import tomllib
from copy import deepcopy
from itertools import islice
from multiprocessing import Pool
from multiprocessing.pool import ThreadPool
from pathlib import Path
from shutil import rmtree
from typing import Any, Callable, List, Union
from urllib.parse import ParseResult, urlparse
from uuid import uuid4

from loguru import logger

# --------------------------------------------------------------------------------------------------


# None Type
NoneType = type(None)


# --------------------------------------------------------------------------------------------------


def isTrue(target: object, typeClass: Any) -> bool:
    """检查对象是否为真"""

    # 常见布尔类型:
    #
    #     Boolean     bool            False
    #     Numbers     int/float       0/0.0
    #     String      str             ""
    #     List        list/tuple/set  []/()/{}
    #     Dictionary  dict            {}
    #
    # 查看变量类型: type(x)
    #
    # 判断变量类型: isinstance(x, str)
    #
    # 函数使用 callable(func) 判断
    #
    # 判断多个类型:
    #
    #   isTrue("abc", (str, int))
    #   isTrue("abc", (str | int))
    #
    # all() 用于检查一个可迭代对象(如列表、元组、集合等)中的 所有 元素是否为 真值 (truthy), 所有元素为真, 返回 True
    # any() 用于检查一个可迭代对象(如列表、元组、集合等)中的 某个 元素是否为 真值 (truthy), 某个元素为真, 返回 True
    # 与 all() 作用相反的 not any(), 可以用来检查所有元素是否为 假值 (falsy), any() 中所有元素为假, not any() 返回 True
    #
    # return target not in [False, None, 0, 0.0, '', (), [], {}, {*()}, {*[]}, {*{}}]

    try:
        return isinstance(target, typeClass) and bool(target)
    except Exception as e:
        logger.exception(e)
        return False


# --------------------------------------------------------------------------------------------------


def check_arguments(data: list) -> bool:
    """检查函数参数"""

    # data 示例: [(name, str, "name"), (age, int, "age")]

    try:
        if not isTrue(data, list):
            logger.error("argument error: data")
            return False
        for arg, arg_type, arg_name in data:
            if not isTrue(arg, arg_type):
                logger.error(f"argument error: {arg_name}")
                return False
        return True
    except Exception as e:
        logger.exception(e)
        return False


# --------------------------------------------------------------------------------------------------


def os_environ(
    name: str,
    value: Any = None
) -> Any:
    """系统变量"""

    # 伪全局变量
    # Python 没有全局变量, 多个文件无法调用同一个变量.
    # 为了解决这个问题, 将变量设置为系统变量, 从而实现多个文件调用同一个变量.

    # 检查参数是否正确
    if not check_arguments([(name, str, "name")]):
        return None

    try:

        # 变量名添加一个前缀, 防止和系统中其它变量名冲突
        _variable_name = f'PYTHON_VARIABLE_{name}'

        # 如果 value 的值是 None, 则从系统环境获取变量数据
        if value is None:

            _data = os.environ.get(_variable_name)

            # 判断是否有数据
            if _data is None or not isTrue(_data, str):
                return None

            # 使用 json.loads() 解析数据
            parsed_data = json.loads(_data)
            return parsed_data

        # 如果 value 的值不是 None, 则保存数据到系统环境变量
        _data = json.dumps(value)
        os.environ[_variable_name] = _data

        return value

    except Exception as e:
        logger.exception(e)
        return None


# --------------------------------------------------------------------------------------------------


def mam_of_numbers(
    numbers: list | tuple,
    dest_type: str | None = None
) -> tuple | None:
    """返回一组数字中的 最大值(maximum), 平均值(average), 最小值(minimum)"""

    # numbers 数字列表 (仅支持 list 和 tuple, 不支 set)
    # dest_type 目标类型 (将数字列表中的数字转换成统一的类型)

    # 检查参数是否正确
    if not check_arguments([(numbers, (list, tuple), "numbers")]):
        return None

    if not any([isTrue(dest_type, str), dest_type is None]):
        logger.error("argument error: dest_type")
        return None

    try:

        _numbers = deepcopy(numbers)

        # 转换数据类型
        if dest_type == "float":
            _numbers = [float(i) for i in numbers]

        if dest_type == "int":
            _numbers = [int(i) for i in numbers]

        # 提取数据
        _num_max = max(_numbers)
        _num_avg = f"{sum(_numbers) / len(_numbers):.2f}"
        _num_min = min(_numbers)

        if dest_type == int:
            _num_avg = int(_num_avg)
        else:
            _num_avg = float(_num_avg)

            # 返回数据
        return _num_max, _num_avg, _num_min

    except Exception as e:
        logger.exception(e)
        return None


# --------------------------------------------------------------------------------------------------


def step_number_for_split_equally(
    integer: int,
    split_equally_number: int
) -> int | None:
    """
    step number for split equally

    平分数字的步长

      integer 数字
      split_equally_number 平分 integer 的数字

    示例:

        [1, 2, 3, 4, 5, 6, 7, 8, 9]

        分成 2 份 -> [[1, 2, 3, 4, 5], [6, 7, 8, 9]] -> 返回 5
        分成 3 份 -> [[1, 2, 3], [4, 5, 6], [7, 8, 9]] -> 返回 3
        分成 4 份 -> [[1, 2, 3], [4, 5], [6, 7], [8, 9]] -> 返回 3
        分成 5 份 -> [[1, 2], [3, 4], [5, 6], [7, 8], [9]] -> 返回 2
    """
    try:
        if integer % split_equally_number == 0:
            return int(integer / split_equally_number)
        return int(integer / split_equally_number) + 1
    except Exception as e:
        logger.exception(e)
        return None


# --------------------------------------------------------------------------------------------------


def division(
    dividend: int | float,
    divisor: int | float
) -> float | None:
    """Division"""
    # 检查参数是否正确
    if not check_arguments([(dividend, (int, float), "dividend"), (divisor, (int, float), "divisor")]):
        return None
    try:
        return dividend / divisor
    except Exception as e:
        logger.exception(e)
        return None


# --------------------------------------------------------------------------------------------------


def check_file_type(
    file_object: str,
    file_type: str
) -> bool:
    """检查文件类型"""

    # file_object 文件对象
    # file_type 文件类型

    # 检查参数是否正确
    if not check_arguments([(file_object, str, "file"), (file_type, str, "file_type")]):
        return False

    try:

        _file_path = Path(file_object)

        match True:
            case True if _file_path.exists() is False:
                result = False
            case True if file_type == 'absolute' and _file_path.is_absolute() is True:
                result = True
            case True if file_type == 'block_device' and _file_path.is_block_device() is True:
                result = True
            case True if file_type == 'dir' and _file_path.is_dir() is True:
                result = True
            case True if file_type == 'fifo' and _file_path.is_fifo() is True:
                result = True
            case True if file_type == 'file' and _file_path.is_file() is True:
                result = True
            case True if file_type == 'mount' and _file_path.is_mount() is True:
                result = True
            case True if file_type == 'relative_to' and _file_path.is_relative_to() is True:
                result = True
            case True if file_type == 'reserved' and _file_path.is_reserved() is True:
                result = True
            case True if file_type == 'socket' and _file_path.is_socket() is True:
                result = True
            case True if file_type == 'symlink' and _file_path.is_symlink() is True:
                result = True
            case _:
                result = False

        return result

    except Exception as e:
        logger.exception(e)
        return False


# --------------------------------------------------------------------------------------------------


def list_sort(
    data: list,
    deduplication: bool = False,
    **kwargs
) -> list | None:
    """list sort"""
    # 列表排序, 示例: list_sort(['1.2.3.4', '2.3.4.5'], key=inet_aton)
    # 参考文档:
    #     https://stackoverflow.com/a/4183538
    #     https://blog.csdn.net/u013541325/article/details/117530957
    try:

        # from ipaddress import ip_address
        # _ips = [str(i) for i in sorted(ip_address(ip.strip()) for ip in ips)]
        # 注意: list.sort() 是直接改变 list, 不会返回 list

        # 拷贝数据, 去重, 排序, 返回
        _data = deepcopy(data)
        if deduplication is True:
            _data = list(set(_data))
        _data.sort(**kwargs)
        return _data

    except Exception as e:
        logger.exception(e)
        return None


def list_dict_sorted_by_key(
    data: list | tuple,
    key: str,
    **kwargs
) -> list | None:
    """list dict sorted by key"""
    # 列表字典排序
    # 参考文档: https://stackoverflow.com/a/73050
    try:
        _data = deepcopy(data)
        return sorted(_data, key=lambda x: x[key], **kwargs)
    except Exception as e:
        logger.exception(e)
        return None

def list_split(
    data: list,
    number: int,
    equally: bool = False
) -> list | None:
    """列表分割"""

    # 列表分割
    #
    # 默认: 将 list 以 number 个元素为一个子 list 分割
    #
    #   data = [1, 2, 3, 4, 5, 6, 7]                                奇数个元素
    #   list_split(data, 2) -> [[1, 2], [3, 4], [5, 6], [7]]        将 data 以 2个元素 为一个 list 分割
    #   list_split(data, 3) -> [[1, 2, 3], [4, 5, 6], [7]]          将 data 以 3个元素 为一个 list 分割
    #
    #   data = [1, 2, 3, 4, 5, 6, 7, 8]                             偶数个元素
    #   list_split(data, 2) -> [[1, 2], [3, 4], [5, 6], [7, 8]]     将 data 以 2个元素 为一个 list 分割
    #   list_split(data, 3) -> [[1, 2, 3], [4, 5, 6], [7, 8]]       将 data 以 3个元素 为一个 list 分割
    #
    # equally 为 True 时, 将 list 平均分成 number 个元素的子 list
    #
    #   data = [1, 2, 3, 4, 5, 6, 7]                                奇数个元素
    #   list_split(data, 2, True) -> [[1, 2, 3, 4], [5, 6, 7]]      将 data 平均分成 2个子list
    #   list_split(data, 3, True) -> [[1, 2, 3], [4, 5, 6], [7]]    将 data 平均分成 3个子list
    #
    #   data = [1, 2, 3, 4, 5, 6, 7, 8]                             偶数个元素
    #   list_split(data, 2, True) -> [[1, 2, 3, 4], [5, 6, 7, 8]]   将 data 平均分成 2个子list
    #   list_split(data, 3, True) -> [[1, 2, 3], [4, 5, 6], [7, 8]] 将 data 平均分成 3个子list

    # 判断参数是否正确
    match True:
        case True if not isTrue(data, list):
            logger.error("argument error: data")
            return None
        case True if not isTrue(number, int):
            logger.error("argument error: number")
            return None
        case _:
            pass

    try:

        # 要将列表平均分成 n 个子列表
        if isTrue(equally, bool):
            it = iter(data)
            chunk_size = (len(data) + number - 1) // number  # 每组至少多少个元素
            return [list(islice(it, chunk_size)) for _ in range(number)]

        # 将列表按每 n 个元素为一个列表进行分割
        it = iter(data)
        return [list(islice(it, number)) for _ in range((len(data) + number - 1) // number)]

    except Exception as e:
        logger.exception(e)
        return None


def list_print_by_step(
    data: list,
    step: int,
    separator: str = " "
) -> bool:
    """根据 步长 和 分隔符 有规律的打印列表中的数据"""

    # 判断参数是否正确
    match True:
        case True if not isTrue(data, list):
            logger.error("argument error: data")
            return False
        case True if not isTrue(step, int):
            logger.error("argument error: step")
            return False
        case True if not isTrue(separator, str):
            logger.error("argument error: separator")
            return False
        case _:
            pass

    try:

        # result: list = []

        # if len(data) <= step:
        #     result.append(data)
        # else:
        #     data_list = list_split(data, step, debug=debug)
        #     if data_list is None or isTrue(data_list, list) is False:
        #         return False
        #     result = data_list

        # for item in result:
        #     print(*item, sep=separator)

        # return True

        # 打印
        for i, v in enumerate(data):
            if i > 0 and i % step == 0:
                print()
            print(v, end=separator)

        print()

        return True

    except Exception as e:
        logger.exception(e)
        return False


def list_remove_list(
    original: list,
    remove: list
) -> list | None:
    """List remove List"""
    try:
        _original = deepcopy(original)
        _remove = deepcopy(remove)
        return [i for i in _original if i not in _remove]
    except Exception as e:
        logger.exception(e)
        return None


def list_merge(
    data: list
) -> list | None:
    """list merge"""
    # 合并 List 中的 List 为一个 List
    try:
        _results = []
        for i in deepcopy(data):
            _results += i
        return _results
    except Exception as e:
        logger.exception(e)
        return None


def list_to_file(
    data: list,
    file: str,
    encoding: str = 'utf-8-sig'
) -> bool:
    """list to file"""
    try:
        with open(file, 'w', encoding=encoding) as _file:
            for line in data:
                _file.write(f"{line}\n")
        return True
    except Exception as e:
        logger.exception(e)
        return False

def list_to_csvfile(
    data: list,
    file: str,
    fields: list | None = None,
    encoding: str = 'utf-8-sig',
    **kwargs
) -> bool:
    """list to csvfile"""
    try:
        with open(file, 'w', encoding=encoding) as _file:
            # CRLF replaced by LF
            # https://stackoverflow.com/a/29976091
            outcsv = csv.writer(_file, lineterminator=os.linesep, **kwargs)
            if fields is not None and isTrue(fields, list):
                outcsv.writerow(fields)
            outcsv.writerows(data)
        return True
    except Exception as e:
        logger.exception(e)
        return False

def range_zfill(
    start: int,
    stop: int,
    step: int,
    width: int
) -> list | None:
    """range zfill"""
    # 生成长度相同的字符串的列表
    # 示例: range_zfill(8, 13, 1, 2) => ['08', '09', '10', '11', '12']
    # 生成 小时 列表: range_zfill(0, 24, 1, 2)
    # 生成 分钟和秒 列表: range_zfill(0, 60, 1, 2)
    # https://stackoverflow.com/a/733478
    # the zfill() method to pad a string with zeros
    try:
        return [str(i).zfill(width) for i in range(start, stop, step)]
    except Exception as e:
        logger.exception(e)
        return None


# --------------------------------------------------------------------------------------------------


def dict_remove_key(
    data: dict,
    key: str
) -> dict | None:
    """dict remove key"""
    try:
        data_copy: dict = deepcopy(data)
        data_copy.pop(key)
        return data_copy
    except Exception as e:
        logger.exception(e)
        return None

def dict_to_file(
    data: dict,
    file: str,
    encoding: str = 'utf-8-sig',
    **kwargs
) -> bool:
    """dict to file"""
    try:
        with open(file, 'w', encoding=encoding) as _file:
            json.dump(obj=data, fp=_file, indent=4, sort_keys=True, **kwargs)
        return True
    except Exception as e:
        logger.exception(e)
        return False


def dict_nested_update(
    data: dict,
    key: str,
    value: Any
) -> bool:
    """dict nested update"""
    # dictionary nested update
    # https://stackoverflow.com/a/58885744
    try:

        if not isTrue(data, dict):
            return False

        for _k, _v in data.items():
            # callable() 判断是非为 function
            if (key is not None and key == _k) or (callable(key) is True and key == _k):
                if callable(value) is True:
                    data[_k] = value()
                else:
                    data[_k] = value
            elif isTrue(_v, dict):
                dict_nested_update(_v, key, value)
            elif isTrue(_v, list):
                for _o in _v:
                    if isTrue(_o, dict):
                        dict_nested_update(_o, key, value)
            else:
                pass

        return True

    except Exception as e:
        logger.exception(e)
        return False


# --------------------------------------------------------------------------------------------------


def filename(
    file: str,
    split: str = '.'
) -> str | None:
    """filename"""
    # 获取文件名称
    # https://stackoverflow.com/questions/678236/how-do-i-get-the-filename-without-the-extension-from-a-path-in-python
    # https://stackoverflow.com/questions/4152963/get-name-of-current-script-in-python
    try:
        _basename = str(os.path.basename(file))
        _index_of_split = _basename.index(split)
        return _basename[:_index_of_split]
    except Exception as e:
        logger.exception(e)
        return None


def filehash(
    file: str,
    sha: str = 'md5'
) -> str | None:
    """filehash"""
    # 获取文件Hash
    # 参考文档:
    #     https://stackoverflow.com/a/59056837
    #     https://stackoverflow.com/questions/22058048/hashing-a-file-in-python
    try:
        with open(file, "rb") as _file:
            match True:
                case True if sha == 'sha1':
                    file_hash = hashlib.sha1()
                case True if sha == 'sha224':
                    file_hash = hashlib.sha224()
                case True if sha == 'sha256':
                    file_hash = hashlib.sha256()
                case True if sha == 'sha384':
                    file_hash = hashlib.sha384()
                case True if sha == 'sha512':
                    file_hash = hashlib.sha512()
                case True if sha == 'sha3_224':
                    file_hash = hashlib.sha3_224()
                case True if sha == 'sha3_256':
                    file_hash = hashlib.sha3_256()
                case True if sha == 'sha3_384':
                    file_hash = hashlib.sha3_384()
                case True if sha == 'sha3_512':
                    file_hash = hashlib.sha3_512()
                # case True if sha == 'shake_128':
                #     file_hash = hashlib.shake_128()
                # case True if sha == 'shake_256':
                #     file_hash = hashlib.shake_256()
                case _:
                    file_hash = hashlib.md5()
            # 建议设置为和 block size 相同的值, 多数系统默认为 4096, 可使用 stat 命令查看
            # stat / (IO Block)
            # stat -f / (Block size)
            while chunk := _file.read(4096):
                file_hash.update(chunk)
            return file_hash.hexdigest()
    except Exception as e:
        logger.exception(e)
        return None


def filesize(
    file: str
) -> int | None:
    """filesize"""
    # 获取文件大小
    try:
        return os.path.getsize(file)
    except Exception as e:
        logger.exception(e)
        return None


# --------------------------------------------------------------------------------------------------


# def resolve_path() -> str | None:
#     """resolve path"""
#     # 获取当前目录名称
#     return str(Path().resolve())


# def parent_path(
#     path: str,
#     debug: bool = False,
#     **kwargs
# ) -> str | None:
#     """获取父目录名称"""
#     try:
#         return str(Path(path, **kwargs).parent.resolve()) if isTrue(path, str, debug=debug) else None
#     except Exception as e:
#         if isTrue(debug, bool):
#             logger.exception(e)
#         return None


def realpath(
    path: str,
    **kwargs
) -> str | None:
    """获取对象真实路径"""
    try:
        if not isTrue(path, str):
            return None
        return str(Path(path, **kwargs).resolve())
    except Exception as e:
        logger.exception(e)
        return None


def current_dir(
    path: str,
    **kwargs
) -> str | None:
    """获取对象所在目录"""
    try:
        if not isTrue(path, str):
            return None
        return str(Path(path, **kwargs).parent.resolve())
    except Exception as e:
        logger.exception(e)
        return None


def parent_dir(
    path: str,
    **kwargs
) -> str | None:
    """获取对象所在目录的父目录"""
    try:
        if not isTrue(path, str):
            return None
        return str(Path(path, **kwargs).parent.parent.resolve())
    except Exception as e:
        logger.exception(e)
        return None


# --------------------------------------------------------------------------------------------------


def retry(func: Callable, times: int = 3, **kwargs):
    """重试"""

    # 函数传递参数: https://stackoverflow.com/a/803632
    # callable() 判断类型是非为函数: https://stackoverflow.com/a/624939

    for attempt in range(times):
        try:
            # 执行函数并结果
            return func(**kwargs)
        except Exception as e:
            logger.exception(e)
            if attempt < (times - 1):
                logger.info('retrying ...')
            else:
                logger.error("all retries failed")
                return False


# --------------------------------------------------------------------------------------------------

# 日期时间有两种: UTC datetime (UTC时区日期时间) 和 Local datetime (当前时区日期时间)
#
# Unix Timestamp 仅为 UTC datetime 的值
#
# 但是, Local datetime 可以直接转换为 Unix Timestamp, UTC datetime 需要先转换到 UTC TimeZone 再转换为 Unix Timestamp
#
# 相反, Unix Timestamp 可以直接转换为 UTC datetime, 要获得 Local datetime, 需要再将 UTC datetime 转换为 Local datetime
#
#     https://stackoverflow.com/a/13287083
#     https://stackoverflow.com/a/466376
#     https://stackoverflow.com/a/7999977
#     https://stackoverflow.com/a/3682808
#     https://stackoverflow.com/a/63920772
#     https://www.geeksforgeeks.org/how-to-remove-timezone-information-from-datetime-object-in-python/
#
# pytz all timezones
#
#     https://stackoverflow.com/a/13867319
#     https://stackoverflow.com/a/15692958
#
#     import pytz
#     pytz.all_timezones
#     pytz.common_timezones
#     pytz.timezone('US/Eastern')
#
# timezone
#
#     https://stackoverflow.com/a/39079819
#     https://stackoverflow.com/a/1681600
#     https://stackoverflow.com/a/4771733
#     https://stackoverflow.com/a/63920772
#     https://toutiao.io/posts/sin4x0/preview
#
# 其它:
#
#     dt.replace(tzinfo=timezone.utc).astimezone(tz=None)
#
#     (dt.replace(tzinfo=timezone.utc).astimezone(tz=None)).strftime(format)
#     datetime.fromisoformat((dt.replace(tzinfo=timezone.utc).astimezone(tz=None)).strftime(format))
#     string_to_datetime((dt.replace(tzinfo=timezone.utc).astimezone(tz=None)).strftime(format), format)
#
#     datetime.fromisoformat(time.strftime(format, time.gmtime(dt)))


def date_to_datetime(
    date_object: datetime.datetime
) -> datetime.datetime | None:
    """'日期'转换为'日期时间'"""
    # https://stackoverflow.com/a/1937636
    try:
        return datetime.datetime.combine(date_object, datetime.datetime.min.time())
    except Exception as e:
        logger.exception(e)
        return None


def datetime_to_date(
    datetime_instance: datetime.datetime
) -> datetime.date | None:
    """'日期时间'转换为'日期'"""
    # https://stackoverflow.com/a/3743240
    try:
        return datetime_instance.date()
    except Exception as e:
        logger.exception(e)
        return None


def local_timezone():
    """获取当前时区"""
    return datetime.datetime.now(datetime.timezone.utc).astimezone().tzinfo


def datetime_now(**kwargs) -> datetime.datetime | None:
    """获取当前日期和时间"""
    utc = kwargs.pop("utc", False)
    try:
        if isTrue(utc, bool):
            return datetime.datetime.now(datetime.timezone.utc)
        return datetime.datetime.now(**kwargs)
    except Exception as e:
        logger.exception(e)
        return None


def datetime_offset(
    datetime_instance: datetime.datetime | None = None,
    **kwargs
) -> datetime.datetime | None:
    """
    获取 '向前或向后特定日期时间' 的日期和时间

    类型: weeks, days, hours, minutes, seconds, microseconds, milliseconds
    """
    _utc = kwargs.pop("utc", False)
    try:
        if isinstance(datetime_instance, datetime.datetime):
            return datetime_instance + datetime.timedelta(**kwargs)

        if _utc is True:
            return datetime.datetime.now(datetime.timezone.utc) + datetime.timedelta(**kwargs)

        return datetime.datetime.now() + datetime.timedelta(**kwargs)

    except Exception as e:
        logger.exception(e)
        return None


def datetime_to_string(
    datetime_instance: datetime.datetime,
    string_format: str = '%Y-%m-%d %H:%M:%S'
) -> str | None:
    """'日期时间'转换为'字符串'"""
    try:
        if not isTrue(datetime_instance, datetime.datetime):
            return None
        return datetime.datetime.strftime(datetime_instance, string_format)
    except Exception as e:
        logger.exception(e)
        return None


def datetime_to_timestamp(
    datetime_instance: datetime.datetime,
    utc: bool = False
) -> int | None:
    """
    Datatime 转换为 Unix Timestamp
    Local datetime 可以直接转换为 Unix Timestamp
    UTC datetime 需要先替换 timezone 再转换为 Unix Timestamp
    """
    try:
        if not isTrue(datetime_instance, datetime.datetime):
            return None
        return int(datetime_instance.replace(tzinfo=datetime.timezone.utc).timestamp()) if utc is True else int(datetime_instance.timestamp())
    except Exception as e:
        logger.exception(e)
        return None


def datetime_local_to_timezone(
    datetime_instance: datetime.datetime,
    tz: datetime.timezone = datetime.timezone.utc
) -> datetime.datetime | None:
    """
    Local datetime to TimeZone datetime(默认转换为 UTC datetime)
    replace(tzinfo=None) 移除结尾的时区信息
    """
    try:
        if not isTrue(datetime_instance, datetime.datetime):
            return None
        return (datetime.datetime.fromtimestamp(datetime_instance.timestamp(), tz=tz)).replace(tzinfo=None)
    except Exception as e:
        logger.exception(e)
        return None


def datetime_utc_to_timezone(
    datetime_instance: datetime.datetime,
    tz: Any = datetime.datetime.now(datetime.timezone.utc).astimezone().tzinfo
) -> datetime.datetime | None:
    """
    UTC datetime to TimeZone datetime(默认转换为 Local datetime)
    replace(tzinfo=None) 移除结尾的时区信息
    """
    try:
        if not isTrue(datetime_instance, datetime.datetime):
            return None
        return datetime_instance.replace(tzinfo=datetime.timezone.utc).astimezone(tz).replace(tzinfo=None)

    except Exception as e:
        logger.exception(e)
        return None


def timestamp_to_datetime(
    timestamp: int,
    tz: datetime.timezone = datetime.timezone.utc
) -> datetime.datetime | None:
    """Unix Timestamp 转换为 Datatime"""
    try:
        if not isTrue(timestamp, int):
            return None
        return (datetime.datetime.fromtimestamp(timestamp, tz=tz)).replace(tzinfo=None)
    except Exception as e:
        logger.exception(e)
        return None


def datetime_string_to_datetime(
    datetime_string: str,
    datetime_format: str = '%Y-%m-%d %H:%M:%S'
) -> datetime.datetime | None:
    """datetime string to datetime"""
    try:
        if not isTrue(datetime_string, str):
            return None
        return datetime.datetime.strptime(datetime_string, datetime_format)
    except Exception as e:
        logger.exception(e)
        return None


def datetime_string_to_timestamp(
    datetime_string: str,
    datetime_format: str = '%Y-%m-%d %H:%M:%S'
) -> int | None:
    """datetime string to timestamp"""
    try:
        if not isTrue(datetime_string, str):
            return None
        return int(time.mktime(time.strptime(datetime_string, datetime_format)))
    except Exception as e:
        logger.exception(e)
        return None


def datetime_object(
    date_time: datetime.datetime
) -> dict | None:
    """datetime object"""
    try:
        return {
            'date': date_time.strftime("%Y-%m-%d"),
            'time': date_time.strftime("%H:%M:%S"),
            'datetime_now': date_time.strftime("%Y-%m-%d %H:%M:%S"),
            'datetime_minute': date_time.strftime("%Y-%m-%d %H:%M:00"),
            'datetime_hour': date_time.strftime("%Y-%m-%d %H:00:00"),
            'datetime_zero': date_time.strftime('%Y-%m-%d 00:00:00')
        }
    except Exception as e:
        logger.exception(e)
        return None


# --------------------------------------------------------------------------------------------------


# run_cmd = bash('echo ok', universal_newlines=True, stdout=PIPE)
#
# if run_cmd != None:
#     returncode = run_cmd.returncode
#     outputs = run_cmd.stdout.splitlines()
#     print(returncode, type(returncode))
#     print(outputs, type(outputs))
#
# # echo 'echo ok' > /tmp/ok.sh
# run_script = bash('/tmp/ok.sh', file=True, universal_newlines=True, stdout=PIPE)
#
# if run_script != None:
#     returncode = run_script.returncode
#     outputs = run_script.stdout.splitlines()
#     print(returncode, type(returncode))
#     print(outputs, type(outputs))


def shell(
    command: str,
    isfile: bool = False,
    sh_shell: str = '/bin/bash',
    sh_option: str | None = None,
    **kwargs
) -> subprocess.CompletedProcess | None:
    """run shell command or script"""
    try:
        match True:
            case True if not check_file_type(sh_shell, 'file'):
                return None
            case True if isTrue(sh_shell, str) and isTrue(command, str):
                if isfile is True:
                    if sh_option is None:
                        return subprocess.run([sh_shell, command], **kwargs, check=False)
                    return subprocess.run([sh_shell, sh_option, command], **kwargs, check=False)
                if sh_option is None:
                    sh_option = '-c'
                return subprocess.run([sh_shell, sh_option, command], **kwargs, check=False)
            case _:
                return None
    except Exception as e:
        logger.exception(e)
        return None


# --------------------------------------------------------------------------------------------------


def json_file_parser(
    file: str
) -> dict | None:
    """JSON File Parser"""
    try:
        if not check_file_type(file, 'file'):
            logger.error(f"no such file: {file}")
            return None
        with open(file, encoding="utf-8") as json_raw:
            json_dict = json.load(json_raw)
        return json_dict
    except Exception as e:
        logger.exception(e)
        return None

# json_raw = '''
# {
#     "markdown.preview.fontSize": 14,
#     "editor.minimap.enabled": false,
#     "workbench.iconTheme": "vscode-icons",
#     "http.proxy": "http://127.0.0.1:1087"
# }
# '''
#
# print(json_sort(json_raw))
#
# {
#     "editor.minimap.enabled": false,
#     "http.proxy": "http://127.0.0.1:1087",
#     "markdown.preview.fontSize": 14,
#     "workbench.iconTheme": "vscode-icons"
# }


def json_sort(
    string: str,
    **kwargs
) -> str | None:
    """JSON Sort"""
    try:
        if not isTrue(string, str):
            return None
        return json.dumps(json.loads(string), indent=4, sort_keys=True, **kwargs)
    except Exception as e:
        logger.exception(e)
        return None


# --------------------------------------------------------------------------------------------------


def delete_files(
    files: Union[str, List]
) -> bool:
    """删除文件"""
    try:

        if isinstance(files, str) and check_file_type(files, 'file'):
            os.remove(files)
            logger.success(f'deleted file: {files}')
            return True

        if not isTrue(files, list):
            return False

        for _file in files:

            if isinstance(_file, str) and check_file_type(_file, 'file'):
                try:
                    os.remove(_file)
                    logger.success(f'deleted file: {_file}')
                except Exception as e:
                    logger.error(f'error file: {_file} {e}')
            else:
                logger.error(f'error file: {_file}')

        return True

    except Exception as e:
        logger.exception(e)
        return False


def delete_directory(
    directory: Union[str, List]
) -> bool:
    """
    delete directory

    https: // docs.python.org / 3 / library / os.html  # os.rmdir

        os.rmdir(path, *, dir_fd=None)

    Remove(delete) the directory path.

    If the directory does not exist or is not empty, an FileNotFoundError or an OSError is raised respectively.

    In order to remove whole directory trees, shutil.rmtree() can be used.

    https: // docs.python.org / 3 / library / shutil.html  # shutil.rmtree

        shutil.rmtree(path, ignore_errors=False, onerror=None)

    Delete an entire directory tree; path must point to a directory(but not a symbolic link to a directory).

    If ignore_errors is true, errors resulting from failed removals will be ignored;

    if false or omitted, such errors are handled by calling a handler specified by onerror or , if that is omitted, they raise an exception.
    """
    try:

        if isinstance(directory, str) and check_file_type(directory, 'dir'):
            rmtree(directory)
            logger.success(f'deleted directory: {directory}')
            return True

        if not isTrue(directory, list):
            logger.error(f'error directory: {directory}')
            return False

        for _dir in directory:

            if isTrue(_dir, str) and check_file_type(_dir, 'dir'):
                try:
                    rmtree(_dir)
                    logger.success(f'deleted directory: {_dir}')
                except Exception as e:
                    logger.error(f'error directory: {_dir} {e}')
            else:
                logger.error(f'error directory: {_dir}')

        return True

    except Exception as e:
        logger.exception(e)
        return False


# --------------------------------------------------------------------------------------------------


def processor(
    process_func: Callable,
    process_data: List[Any],
    process_num: int = 2,
    thread: bool = False,
    **kwargs
) -> Union[List[Any], bool]:
    """使用多线程或多进程对数据进行并行处理"""

    # :param process_func: 处理函数
    # :param process_data: 待处理数据列表
    # :param process_num: 并行数量
    # :param thread: 是否使用多线程
    # :param kwargs: 其他可选参数传递给线程池或进程池
    # :return: 处理后的结果列表或 False（异常情况）
    #
    # MultiThread   多线程
    # MultiProcess  多进程
    #
    # ThreadPool    线程池
    # Pool          进程池
    #
    # ThreadPool 共享内存, Pool 不共享内存
    # ThreadPool 可以解决 Pool 在某些情况下产生的 Can't pickle local object 的错误
    #   https://stackoverflow.com/a/58897266
    #
    # 如果要启动一个新的进程或者线程, 将 process_num 设置为 1 即可

    try:

        # 检查参数
        if not check_arguments([(process_data, list, "process_data")]):
            return False

        # 确保并行数不超过数据量
        process_num = min(len(process_data), process_num)
        _data_chunks = (
            list_split(process_data, process_num, equally=True)
            if process_num > 1
            else [process_data]
        )

        if not _data_chunks:
            logger.error("data chunks error")
            return False

        logger.info(
            f"Starting {'multi-threading' if thread else 'multi-processing'} with {process_num} workers..."
        )

        # 执行多线程或多进程任务
        pool_cls = ThreadPool if thread else Pool
        with pool_cls(process_num, **kwargs) as pool:
            return pool.map(process_func, _data_chunks)

    except Exception as e:
        logger.exception(e)
        return False


# --------------------------------------------------------------------------------------------------


def create_empty_file(
    file: str | None = None
) -> str | None:
    """create empty file"""
    try:
        if file is None:
            # 当前时间戳(纳秒)
            timestamp = time.time_ns()
            # 空文件路径
            file = f'/tmp/none_{timestamp}.txt'
        # 创建一个空文件
        # pylint: disable=R1732
        open(file, "w", encoding="utf-8").close()
        # 返回文件路径
        return file
    except Exception as e:
        logger.exception(e)
        return None


# --------------------------------------------------------------------------------------------------


def uuid4_hex() -> str:
    """UUID"""
    return uuid4().hex


def increment_version(
    version: str
) -> str | None:
    """版本号递增"""
    try:
        version_numbers = version.split('.')
        version_numbers[-1] = str(int(version_numbers[-1]) + 1)
        return '.'.join(version_numbers)
    except Exception as e:
        logger.exception(e)
        return None


# --------------------------------------------------------------------------------------------------


def make_directory(
    directory: str
) -> bool:
    """创建目录"""
    try:
        os.makedirs(directory)
        return True
    except Exception as e:
        logger.exception(e)
        return False

def change_directory(
    directory: str
) -> bool:
    """改变目录"""
    try:

        if not isTrue(directory, str):
            return False

        if not check_file_type(directory, 'dir'):
            logger.error(f"no such directory: {directory}")
            return False

        os.chdir(directory)

        return True

    except Exception as e:
        logger.exception(e)
        return False


# --------------------------------------------------------------------------------------------------


def load_toml_file(
    file: str
) -> dict | None:
    """Load TOML file"""
    info = '解析配置文件'
    try:
        logger.info(f'{info}[执行]')
        with open(file, "rb") as _file:
            config = tomllib.load(_file)
        logger.success(f'{info}[成功]')
        return config
    except Exception as e:
        logger.error(f'{info}[失败]')
        logger.exception(e)
        return None


def git_clone(
    git_repository: str,
    local_repository: str,
    timeout: int = 30,
    delete: bool = False,
    log_prefix: str = ''
) -> bool:
    """GIT Clone"""

    # 日志前缀
    log_prefix = f'{log_prefix} [GitClone]'

    try:

        # 获取应用程序Git仓库
        # logger.info(f'{log_prefix}process the request')
        # logger.info(f'{log_prefix}git repository: {git_repository}')
        # logger.info(f'{log_prefix}local repository: {local_repository}')

        # 删除本地仓库
        if isTrue(delete, bool):
            delete_directory(local_repository)
            time.sleep(1)

        # from shutil import which
        # logger.info(which('timeout')) if isTrue(debug, bool) else next
        # if which('timeout') != None:
        #     command = f'timeout -s 9 {timeout} git clone {git_repository} {local_repository}'

        # 克隆仓库
        result = shell(
            command=f'timeout -s 9 {timeout} git clone {git_repository} {local_repository}',
            universal_newlines=True,
            stdout=subprocess.PIPE,
            stderr=subprocess.STDOUT
        )

        if result is None:
            return False

        result_code: int = result.returncode
        result_info = result.stdout.splitlines()

        if result_code != 0:
            for i in result_info:
                logger.error(f'{log_prefix}{i}')
            return False

        return True

    except Exception as e:
        logger.error(f'{log_prefix} [failure]')
        logger.exception(e)
        return False


def url_parse(
    url: str,
    scheme: str = 'http'
) -> ParseResult:
    """URL Parse"""
    none_result = ParseResult(scheme='', netloc='', path='', params='', query='', fragment='')
    try:
        # 如果没有 scheme 的话, 字符串是不解析的. 所以, 如果没有 scheme, 就添加一个 scheme, 默认添加 http
        if isTrue(url, str) and (url.find('://') == -1) and isTrue(scheme, str):
            url = f'{scheme}://{url}'
        if isTrue(url, str):
            return urlparse(url)
        return none_result
    except Exception as e:
        logger.exception(e)
        return none_result

# def debug_log(
#     log: None | str = None,
#     exception: None | Exception = None,
#     debug: bool = False,
#     error: bool = False
# ):
#     """debug log"""
#     if isTrue(log, str) and isTrue(debug, bool):
#         if isTrue(error, bool):
#             logger.error(log)
#         else:
#             logger.info(log)
#         return

#     if isTrue(exception, Exception):
#         if isTrue(debug, bool):
#             logger.exception(exception)
#         else:
#             logger.error(exception)



"""Database"""
# Column, Table, MetaData API
#     https://docs.sqlalchemy.org/en/14/core/metadata.html#column-table-metadata-api
# CursorResult
#     https://docs.sqlalchemy.org/en/20/core/connections.html#sqlalchemy.engine.CursorResult
# PostgreSQL 14 Data Types
#     https://www.postgresql.org/docs/14/datatype.html
import csv
from typing import Any

from loguru import logger
from sqlalchemy import CursorResult, Index, create_engine, text
from sqlalchemy.orm import DeclarativeBase

from . import utils


class Database():
    """Database"""

    engine = create_engine('sqlite://')

    def __init__(self, target: str | None = None, **options):
        """Initiation"""
        if isinstance(target, str) and utils.isTrue(target, str):
            if utils.isTrue(options, dict):
                self.engine = create_engine(target, **options)
            else:
                self.engine = create_engine(target)
        else:
            pass

    def initializer(self):
        """ensure the parent proc's database connections are not touched in the new connection pool"""
        self.engine.dispose(close=False)

    def connect_test(self) -> bool:
        info = "Database connect test"
        try:
            logger.info(f"{info} ......")
            self.engine.connect()
            logger.success(f"{info} [success]")
            return True
        except Exception as e:
            logger.error(f"{info} [failure]")
            logger.exception(e)
            return False

    def metadata_init(self, base: DeclarativeBase, **kwargs) -> bool:
        # https://stackoverflow.com/questions/19175311/how-to-create-only-one-table-with-sqlalchemy
        info = "Database init table"
        try:
            logger.info(f"{info} ......")
            base.metadata.drop_all(self.engine, **kwargs)
            base.metadata.create_all(self.engine, **kwargs)
            logger.success(f"{info} [success]")
            return True
        except Exception as e:
            logger.error(f"{info} [failure]")
            logger.exception(e)
            return False

    def create_index(self, index_name, table_field) -> bool:
        # 创建索引
        #   https://stackoverflow.com/a/41254430
        # 示例:
        #   index_name: a_share_list_code_idx1
        #   table_field: Table_a_share_list.code
        info = "Database create index"
        try:
            logger.info(f"{info} ......")
            idx = Index(index_name, table_field)
            try:
                idx.drop(bind=self.engine)
            except Exception as e:
                logger.exception(e)
            idx.create(bind=self.engine)
            logger.success(f'{info} [success]')
            return True
        except Exception as e:
            logger.error(f'{info} [failure]')
            logger.error(e)
            return False

    # 私有函数, 保存 execute 的结果到 CSV 文件
    def _result_save(self, file, data) -> bool:
        try:
            outcsv = csv.writer(file)
            outcsv.writerow(data.keys())
            outcsv.writerows(data)
            return True
        except Exception as e:
            logger.exception(e)
            return False

    def execute(
        self,
        sql: str | None = None,
        sql_file: str | None = None,
        sql_file_kwargs: dict | None = None,
        csv_file: str | None = None,
        csv_file_kwargs: dict | None = None
    ) -> CursorResult[Any] | bool:
        """"运行"""

        # ------------------------------------------------------------

        # 提取 SQL
        # 如果 sql 和 sql_file 同时存在, 优先执行 sql

        sql_object = None

        info: str = f"""Extract SQL: {sql}"""

        try:

            logger.info(f"{info} ......")

            if utils.isTrue(sql, str):

                sql_object = sql

            elif sql_file is not None and utils.isTrue(sql_file, str):

                # 判断文件是否存在
                if isinstance(sql_file, str) and utils.check_file_type(sql_file, "file") is False:

                    logger.error(f"No such file: {sql_file}")
                    return False

                if isinstance(sql_file, str) and utils.isTrue(sql_file, str):

                    # 读取文件内容
                    if sql_file_kwargs is not None and utils.isTrue(sql_file_kwargs, dict):
                        with open(sql_file, "r", encoding="utf-8", **sql_file_kwargs) as _file:
                            sql_object = _file.read()
                    else:
                        with open(sql_file, "r", encoding="utf-8") as _file:
                            sql_object = _file.read()

            else:

                logger.error("SQL or SQL file error")
                logger.error(f"{info} [failure]")
                return False

            logger.success(f'{info} [success]')

        except Exception as e:

            logger.error(f"{info} [failure]")
            logger.exception(e)
            return False

        # ------------------------------------------------------------

        # 执行 SQL

        info = f"""Execute SQL: {sql_object}"""

        try:

            logger.info(f"{info} ......")

            with self.engine.connect() as connect:

                # 执行SQL
                if sql_object is None:
                    return False

                result = connect.execute(text(sql_object))

                connect.commit()

                if csv_file is None:
                    # 如果 csv_file 没有定义, 则直接返回结果
                    logger.success(f'{info} [success]')
                    return result

                # 如果 csv_file 有定义, 则保存结果到 csv_file
                info_of_save = f"Save result to file: {csv_file}"
                logger.info(f"{info_of_save} .......")

                # 保存结果
                if isinstance(csv_file_kwargs, dict) and utils.isTrue(csv_file_kwargs, dict):
                    with open(csv_file, "w", encoding="utf-8", **csv_file_kwargs) as _file:
                        result_of_save = self._result_save(_file, result)
                else:
                    with open(csv_file, "w", encoding="utf-8") as _file:
                        result_of_save = self._result_save(_file, result)

                # 检查保存结果
                if result_of_save is True:
                    logger.success(f'{info_of_save} [success]')
                    logger.success(f'{info} [success]')
                    return True

                logger.error(f"{info_of_save} [failure]")
                logger.error(f"{info} [failure]")
                return False

        except Exception as e:

            logger.error(f'{info} [failure]')
            logger.exception(e)
            return False



"""MongoDB"""
from loguru import logger
from pymongo import MongoClient
from pymongo.collection import Collection

from . import utils


class Mongo():
    """MongoDB"""

    client = MongoClient()

    def __init__(self, target: str | dict | None = None):
        """Initiation"""
        if isinstance(target, str) and utils.isTrue(target, str):
            self.client = MongoClient(target)
        elif isinstance(target, dict) and utils.isTrue(target, dict):
            self.client = MongoClient(**target)
        else:
            pass

    def close(self) -> bool:
        """client close"""
        try:
            self.client.close()
            return True
        except Exception as e:
            logger.exception(e)
            return False

    def connect_test(self) -> bool:
        info = "MongoDB connect test"
        try:
            logger.info(f"{info} ......")
            self.client.server_info()
            logger.success(f"{info} [success]")
            return True
        except Exception as e:
            logger.error(f"{info} [failure]")
            logger.exception(e)
            return False

    def collection(self, database: str, name: str) -> Collection | None:
        try:
            return self.client[database][name]
        except Exception as e:
            logger.exception(e)
            return None

    def collection_insert(self, database, collection, data, drop: bool = False):
        db_collection = self.client[database][collection]
        info = "MongoDB collection insert"
        try:
            logger.info(f"{info} ......")
            # 是否删除 collection
            if utils.isTrue(drop, bool):
                # 删除 collection
                db_collection.drop()
            # 插入数据
            if utils.isTrue(data, dict):
                # 插入一条数据
                result = db_collection.insert_one(data)
            elif utils.isTrue(data, list):
                # 插入多条数据
                result = db_collection.insert_many(data)
            else:
                logger.error(f"{info} [failure]")
                logger.error("Data type error")
                return False
            logger.success(f"{info} [success]")
            return result
        except Exception as e:
            logger.error(f"{info} [failure]")
            logger.exception(e)
            return False



"""HTTP"""
import json
from typing import Any

import requests
from loguru import logger

from . import utils


def download(
    request: dict,
    file: dict,
    chunks: bool = False,
    iter_content: dict | None = None,
    info: str | None = None
) -> bool:
    """下载文件"""

    if utils.isTrue(request, dict):
        request_arguments = {"method": "GET", "stream": True, **request}
    else:
        return False

    if utils.isTrue(file, dict):
        file_arguments = {"mode": "wb", **file}
    else:
        return False

    if iter_content is not None and utils.isTrue(iter_content, dict):
        iter_content_arguments = {"chunk_size": 1024, **iter_content}
    else:
        iter_content_arguments = {"chunk_size": 1024}

    info_prefix: str = "Download"
    if utils.isTrue(info, str):
        info_prefix = f"Download {info}"

    try:

        logger.info(f'{info_prefix} ......')

        response = requests.request(**request_arguments)

        # # pylint: disable=W1514
        with open(**file_arguments) as _file:  # type: ignore

            if utils.isTrue(chunks, bool):
                for _chunk in response.iter_content(**iter_content_arguments):  # type: ignore
                    _file.write(_chunk)
            else:
                _file.write(response.content)

        logger.success(f'{info_prefix} [success]')

        return True

    except Exception as e:

        logger.error(f'{info_prefix} [failure]')
        logger.exception(e)
        return False


def response_json(data: Any = None, **kwargs) -> str | None:
    """解决字符编码问题: ensure_ascii=False"""
    try:
        return json.dumps(data, default=str, ensure_ascii=False, sort_keys=True, **kwargs)
    except Exception as e:
        logger.exception(e)
        return None



"""Redis"""
import redis as RedisClient
from loguru import logger

from . import utils


class Redis:
    """Redis"""

    # https://redis.readthedocs.io/en/stable/_modules/redis/client.html#Redis
    # https://github.com/redis/redis-py#client-classes-redis-and-strictredis
    # redis-py 3.0 drops support for the legacy Redis client class.
    # StrictRedis has been renamed to Redis and an alias named StrictRedis is provided so that users previously using StrictRedis can continue to run unchanged.
    # redis-py 3.0 之后只有一个 Redis, StrictRedis 是 Redis 的别名
    # 这里修改以下参数: host, port, socket_timeout, socket_connect_timeout, charset
    redis = RedisClient.Redis()

    def __init__(self, target: str | dict | None = None):
        """Initiation"""
        if isinstance(target, str) and utils.isTrue(target, str):
            self.redis = RedisClient.from_url(target)
        elif isinstance(target, dict) and utils.isTrue(target, dict):
            self.redis = RedisClient.Redis(**target)
        else:
            pass

    def connect_test(self) -> bool:
        info = "Redis connect test"
        try:
            logger.info(f"{info} ......")
            self.redis.ping()
            logger.success(f"{info} [success]")
            return True
        except Exception as e:
            logger.error(f"{info} [failure]")
            logger.exception(e)
            return False

    def flush(self, flushall: bool = False) -> bool:
        info = "Redis flush"
        try:
            if utils.isTrue(flushall, bool):
                logger.info(f"{info} all ......")
                self.redis.flushall()
            else:
                logger.info(f"{info} db ......")
                self.redis.flushdb()
            logger.success(f"{info} [success]")
            return True
        except Exception as e:
            logger.error(f"{info} [failure]")
            logger.exception(e)
            return False



"""setup"""
from setuptools import find_packages, setup

setup(
    name='ezKit',
    version='1.10.1',
    author='septvean',
    author_email='septvean@gmail.com',
    description='Easy Kit',
    packages=find_packages(exclude=['documents', 'tests']),
    include_package_data=True,
    python_requires='>=3.11',
    install_requires=[
        "loguru>=0.7"
    ]
)
